package edu.uci.ics.seal.savasana;


import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import org.aspectj.lang.Signature;
import org.reflections.Reflections;

import edu.uci.ics.seal.savasana.benchmark.SystemUnderTest;
import edu.uci.ics.seal.savasana.benchmark.annotations.AdaptableComponent;
import edu.uci.ics.seal.savasana.runtime.Adaptor;

public class Util {

	public static void addDelay() {
		try {
			Thread.sleep((long) (Math.random() * Const.AVERAGE_RUNTIME));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Generated trace of program in CSV format. Should be called by all
	 * components to generate standard output.
	 * 
	 * @param component
	 * @param message
	 */
	public synchronized static void log(Class<?> caller, String message, Const.LogLevel logLevel, boolean info) {
		if (logLevel.val >= Const.VERBOSE_LEVEL.val) {
			if (info)
				System.out.println(System.currentTimeMillis() + " , " + caller.getSimpleName() + " , " + message);
			else
				System.out.println(message);
		}
	}

	public synchronized static void log(Class<?> caller, String message, boolean info) {
		log(caller, message, Const.LogLevel.CONCISE, info);
	}

	public synchronized static void log(Class<?> caller, String message, Const.LogLevel logLevel) {
		log(caller, message, logLevel, true);
	}

	public synchronized static void log(Class<?> caller, String message) {
		log(caller, message, true);
	}

	/**
	 * @param thread
	 * @return true if the given thread belongs to the system under test
	 *         threads.
	 */
	public static boolean isSystemUnderTestThread(Thread thread) {
		return thread.getName().toUpperCase().startsWith(Const.SYS_THREAD_PREFIX);
	}

	public static String generateSystemUnderTestTreadName(SystemUnderTest sys, int index) {
		return Const.SYS_THREAD_PREFIX + getSystemUnderTestShortName(sys) + "_" + index;
	}

	public static String getSystemUnderTestShortName(SystemUnderTest sys) {
		String sysName = sys.getClass().getPackage().getName();
		return sysName.substring(sysName.lastIndexOf(".") + 1);
	}

	/**
	 * @return a collection of all threads belongs to the system under test
	 */
	public static List<Thread> getAllSystemUnderTestThreads() {
		return Thread.getAllStackTraces().keySet().stream().filter(Util::isSystemUnderTestThread).sorted((t1, t2) -> t1.getName().compareTo(t2.getName()))
				.collect(Collectors.toList());
	}

	/**
	 * @param signature
	 * @return Find the system under test that matches to the given component's
	 *         signature.
	 */
	public static SystemUnderTest findSysUnderTest(Signature signature) {
		String pkg = signature.getDeclaringType().getPackage().getName();
		return findSysUnderTest(pkg);
	}

	public static SystemUnderTest findSysUnderTest(String pkg) {
		return Arrays.asList(Const.SUTs).stream().filter(x -> x.getClass().getPackage().getName().equals(pkg)).findAny().orElse(null);
	}

	/**
	 * @return true if all no system under test is running.
	 */
	public static boolean checkForTermination() {
		return getAllSystemUnderTestThreads().isEmpty() && Adaptor.resultsPrinted;
	}

	public static String getComponentName(Signature signature, SystemUnderTest sys) {
		Class<?> declaringType = signature.getDeclaringType();
		return getComponentName(declaringType, signature.getDeclaringTypeName(), sys);
	}

	public static String getAdaptableComponent(SystemUnderTest sys) {
		String defaultName = "N/A";
		Reflections reflections = new Reflections(sys.getClass().getPackage().getName());
		Object[] annotated = reflections.getTypesAnnotatedWith(AdaptableComponent.class).toArray();
		if (annotated.length == 0)
			return defaultName;
		return getComponentName((Class<?>) annotated[0], defaultName, sys);
	}

	private static String getComponentName(Class<?> clazz, String defaultName, SystemUnderTest sys) {
		AdaptableComponent annotation = clazz.getAnnotation(AdaptableComponent.class);
		if (annotation == null)
			return defaultName;
		return sys.getClass().getPackage().getName() + "." + annotation.componentName();
	}
	
	public static Long[] getPoissonDistArrivalTime(double lambda, int threadLimit) {
		Long[] waitTimes = new Long[threadLimit];
		waitTimes[0] = (long) 1000;
		for (int i = 1; i < threadLimit; i++) {
			long interArrival = (long) ((-Math.log(1.0 - Math.random()) / lambda) * 1000);
			waitTimes[i] = interArrival;
		}
		return waitTimes;
	}

}
