package edu.uci.ics.seal.savasana.runtime;

import static edu.uci.ics.seal.savasana.Const.AdaptationApproach.TRANQUILITY;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import edu.uci.ics.seal.savasana.Const;
import edu.uci.ics.seal.savasana.Util;
import edu.uci.ics.seal.savasana.benchmark.SystemUnderTest;
import edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.IntervalPlace;

public class Adaptor {

	public static boolean resultsPrinted = false;

	private static Set<String> requestedForAdaptationComponent = new HashSet<String>();

	/**
	 * Records the request for adapting the given component.
	 * 
	 * @param componentId
	 */
	public static void requestForAdaptation(String componentId) {
		if (requestedForAdaptationComponent.contains(componentId))
			return;
		Util.log(Adaptor.class, "Request for adaptation " + componentId, true);
		requestedForAdaptationComponent.add(componentId);
		Registry.logRequestForAdaptation(componentId);
	}

	/**
	 * Suspends the thread when it is in a safe state (being in the safe state
	 * is determined by Monitor according to the selected algorithm). Also the
	 * state of the safe thread (i.e. safe) is recorded in the Registry to be
	 * used later by {@link #checkForComponentsUpdate(String, Thread)
	 * checkForComponentsUpdate}.
	 * 
	 * @param componentID
	 * @param currentThread
	 */
	@SuppressWarnings("deprecation")
	public static void suspendSafeTransactions(String componentID, Thread currentThread) {
		if (!requestedForAdaptationComponent.contains(componentID))
			return;
		if (Registry.isSafeToUpdate(componentID, currentThread)) {
			Util.log(Adaptor.class, "Thread " + currentThread.getName() + " suspended");
			Registry.logThreadSuspension(currentThread);
			currentThread.suspend();
		}
	}

	/**
	 * Adaptor thread periodically checks the registry to see if all threads are
	 * in a safe state for component adaptation
	 */
	public static void checkForComponentsUpdate() {
		for (String componentId : requestedForAdaptationComponent) {
			boolean safeToUpdate = true;
			List<Thread> threadSet = Util.getAllSystemUnderTestThreads();
			for (Thread thread : threadSet)
				if (!Registry.isSafeToUpdate(componentId, thread))
					safeToUpdate = false;
			if (safeToUpdate) {
				// Tranquility suspends threads just before adaptation.
				if (Const.ADAPTATION_APPROACH == TRANQUILITY)
					for (Thread thread : threadSet)
						Adaptor.suspendSafeTransactions(componentId, thread);
				Util.log(Adaptor.class, "Safe to update component " + componentId);
				Util.log(Adaptor.class, getStatusOfRunningThreads());
				Util.log(Adaptor.class, Registry.registeryToString(componentId));
				Registry.logSuccessfulAdaptation(componentId, threadSet);
				updateComponent(componentId);
				resetComponentsState(componentId);
				resumeAllOtherThreads();
				Registry.printExprimentsResult();
				resultsPrinted = true;
				if (Const.EARLY_SHUTDOWN)
					System.exit(0);
			}
		}
	}

	private static void updateComponent(String componentID) {
		Util.log(Adaptor.class, "Replacing component " + componentID + " with a new version");
		SystemUnderTest sys = Util.findSysUnderTest(componentID.substring(0, componentID.lastIndexOf('.')));
		if (sys != null)
			sys.updateComponent();
	}

	private static void resetComponentsState(String componentID) {
		Util.log(Adaptor.class, "Resetting components states");
		for (Thread thread : Util.getAllSystemUnderTestThreads()) {
			Registry.updateRegistry(componentID, thread, "n/a", IntervalPlace.NA, false);
		}
		requestedForAdaptationComponent.remove(componentID);
	}

	@SuppressWarnings("deprecation")
	private static void resumeAllOtherThreads() {
		Util.log(Adaptor.class, "Resuming all suspended threads");
		Util.getAllSystemUnderTestThreads().forEach(Thread::resume);
	}

	private static String getStatusOfRunningThreads() {
		return "Running Threads: " + Util.getAllSystemUnderTestThreads().stream().map(t -> t.getName()).sorted().reduce((t, u) -> t + "," + u).orElse("");
	}
}
