package edu.uci.ics.seal.savasana.runtime_old;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import edu.uci.ics.seal.savasana.benchmark.systemUnderTestInterfaces.ComponentBasedSystemManager;
import edu.uci.ics.seal.savasana.runtime_old.Main_Old.AdaptationApproach;
import edu.uci.ics.seal.savasana.runtime_old.Main_Old.ExperimentMetric;

public class Adaptor implements Runnable {

	private boolean requestForAdaptationFlag;
	private Map<Long, Boolean> currentUnitState;
	private ComponentBasedSystemManager CBSMngr;
	private static Adaptor instance;
	public final static String COMPONENT_THREAD_PREFIX = "ExperimentalThread_";
	private AdaptationApproach approach;
	public static org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger(Adaptor.class.getName());

	private Adaptor() {

	}

	public static Adaptor getInstance(ComponentBasedSystemManager CBSMngr, AdaptationApproach approach) {
		if (instance == null) {
			instance = (Adaptor) Main_Old.context.getBean("adaptor");
			instance.CBSMngr = CBSMngr;
		}
		instance.requestForAdaptationFlag = false;
		instance.approach = approach;
		instance.currentUnitState = new HashMap<Long, Boolean>();
		return instance;
	}

	public void requestForAdaptation() {
		if (Main_Old.metric == ExperimentMetric.REACHABILITY || Main_Old.metric == ExperimentMetric.DISRUPTION)
			log.fatal("S:" + (System.currentTimeMillis() - Main_Old.startTime) + " ");
		System.out.println("Adapter: ######################### Interrupt:Request for Adaptation");
		requestForAdaptationFlag = true;
	}

	@SuppressWarnings("deprecation")
	public void safeToAdapt() {
		Long threadId = Thread.currentThread().getId();
		if (requestForAdaptationFlag) {
			System.out.print("Monitor: *******************");
			System.out.println("Thread " + Thread.currentThread().getName() + " suspended");
			currentUnitState.put(threadId, true);
			if (Main_Old.metric == ExperimentMetric.REACHABILITY || Main_Old.metric == ExperimentMetric.DISRUPTION)
				log.fatal((System.currentTimeMillis() - Main_Old.startTime) + ":");
			Thread.currentThread().suspend();
		}
	}

	public void adaptComponent() {
		if (approach == AdaptationApproach.SERENITY)
			adaptComponentBySerenity();
		else if (approach == AdaptationApproach.TRANQUILITY)
			adaptComponentByTranquility();
	}

	private void adaptComponentBySerenity() {
		while (!isAllOtherThreadsSuspended() && !isExperimentTimeOver()) {
		}
		if (!isExperimentTimeOver()) {
			CBSMngr.adaptComponent();
			log.fatal("E:" + (System.currentTimeMillis() - Main_Old.startTime) + "\n");
			System.out.println("Adapter: ######################### Adaptation Done!");
		}
		requestForAdaptationFlag = false;

		resumeAllOtherThreads();
		//System.exit(0);
	}

	private void adaptComponentByTranquility() {
		while (!isAllThreadsTerminated() && !isExperimentTimeOver()) {
		}
		if (!isExperimentTimeOver()) {
			CBSMngr.adaptComponent();
			log.fatal("E:" + (System.currentTimeMillis() - Main_Old.startTime) + "\n");
			System.out.println("Adapter: ######################### Adaptation Done!");
		}
		//System.exit(0);
	}

	private boolean isExperimentTimeOver() {
		long currentTime = System.currentTimeMillis();
		long period = currentTime - Main_Old.startTime;
		if (period >= Main_Old.EXPERIMENT_TIME)
			return true;
		return false;
	}

	private boolean isAllThreadsTerminated() {
		Set<Thread> threadSet = Thread.getAllStackTraces().keySet();
		for (Thread thread : threadSet) {
			if (thread.getName().startsWith(COMPONENT_THREAD_PREFIX) && Main_Old.runningThreads.contains(thread.getName()))
				return false;
		}
		return true;
	}

	private boolean isAllOtherThreadsSuspended() {
		Collection<Boolean> states = currentUnitState.values();
		int num = 0;
		Set<Thread> threadSet = Thread.getAllStackTraces().keySet();
		for (Thread thread : threadSet) {
			if (thread.getName().startsWith(COMPONENT_THREAD_PREFIX) && Main_Old.runningThreads.contains(thread.getName())) {
				num++;
			}
		}
		num -= states.size();
		if (num == 0)
			return true;
		return false;
	}

	@SuppressWarnings("deprecation")
	private void resumeAllOtherThreads() {
		Set<Thread> threadSet = Thread.getAllStackTraces().keySet();
		for (Thread thread : threadSet) {
			if (!thread.getName().startsWith(COMPONENT_THREAD_PREFIX))
				continue;
			thread.resume();
		}
		Set<Long> keySet = currentUnitState.keySet();
		for (Long key : keySet) {
			currentUnitState.put(key, false);
		}
	}

	@Override
	public void run() {
		try {
			Thread.sleep(Main_Old.WARM_UP_TIME);
			requestForAdaptation();
			adaptComponent();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
