package edu.uci.ics.seal.savasana.runtime_old;

import java.util.ArrayList;
import java.util.List;

import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import edu.uci.ics.seal.savasana.benchmark.systemUnderTestInterfaces.ComponentBasedSystemManager;
import edu.uci.ics.seal.savasana.benchmark.testset.eds.HQUI;

public class Main_Old {

	public static final ApplicationContext context = new ClassPathXmlApplicationContext("springBeanDef.xml");
	public static int EXPERIMENT_TIME;// = 180000;
	public static int WARM_UP_TIME;// = 30000;
	public static long DELAY_UNIT;// = 1000;
	public static long startTime;
	public static long lastRequestArrivalTime;
	public static long[] interArrivalTime;
	public static List<String> runningThreads;
	public static Adaptor adaptor;
	public static TestScenario scenario;
	public static ExperimentMetric metric;
	public static boolean RandomServiceDemand;

	public static void main(String[] args) {
		if (args.length != 10 || !args[3].matches("\\d+(\\.\\d+)*")) {
			System.err
					.println("Provide 10 args: \n1-Mode:staticAnalysi/runtimeContorl \n2-Approach:serenity/tranquility \n3-Scenario \n4-lambda value \n5-thread numbers \n6-warm-up time (request for adaptation time) \n7-service demand time (thread's work time) \n8-random service demand flag \n9-experiment time frame \n10-experiment metric:reachability/disruption");
			return;
		}
		SerenityCycle mode = null;
		AdaptationApproach approach = null;
		if (SerenityCycle.StaticAnalysis.mode.equalsIgnoreCase(args[0]))
			mode = SerenityCycle.StaticAnalysis;
		else if (SerenityCycle.RunTimeContol.mode.equalsIgnoreCase(args[0]))
			mode = SerenityCycle.RunTimeContol;

		if (AdaptationApproach.SERENITY.approach.equalsIgnoreCase(args[1]))
			approach = AdaptationApproach.SERENITY;
		else if (AdaptationApproach.TRANQUILITY.approach.equalsIgnoreCase(args[1]))
			approach = AdaptationApproach.TRANQUILITY;

		if (TestScenario.ANALYZER.name.equalsIgnoreCase(args[2]))
			scenario = TestScenario.ANALYZER;
		else if (TestScenario.ADVISOR.name.equalsIgnoreCase(args[2]))
			scenario = TestScenario.ADVISOR;
		else if (TestScenario.SIMULATOR.name.equalsIgnoreCase(args[2]))
			scenario = TestScenario.SIMULATOR;

		double lambda = Double.parseDouble(args[3]);
		int threadLimit = Integer.parseInt(args[4]);
		WARM_UP_TIME = Integer.parseInt(args[5]);
		DELAY_UNIT = Integer.parseInt(args[6]);
		RandomServiceDemand = Boolean.parseBoolean(args[7]);
		EXPERIMENT_TIME = Integer.parseInt(args[8]);

		if (ExperimentMetric.REACHABILITY.name.equalsIgnoreCase(args[9]))
			metric = ExperimentMetric.REACHABILITY;
		else if (ExperimentMetric.DISRUPTION.name.equalsIgnoreCase(args[9]))
			metric = ExperimentMetric.DISRUPTION;
		else if (ExperimentMetric.OFF.name.equalsIgnoreCase(args[9]))
			metric = ExperimentMetric.OFF;

		try {
			runExperiment(mode, approach, lambda, threadLimit);
		} catch (Exception e) {
			e.printStackTrace();
		}
		JVMShutdownHook hook = new JVMShutdownHook();
		Runtime.getRuntime().addShutdownHook(hook);
	}

	private static void runExperiment(SerenityCycle mode, AdaptationApproach approach, double lambda, int threadLimit)
			throws Exception {
		ComponentBasedSystemManager systemManager = (ComponentBasedSystemManager) context.getBean("edsMain");
		HQUI hqui = init(systemManager, mode, approach, lambda, threadLimit);
		if (mode == SerenityCycle.RunTimeContol && approach == AdaptationApproach.TRANQUILITY
				&& metric != ExperimentMetric.OFF) {
			Thread adaptorThread = new Thread(adaptor);
			adaptorThread.setName("Adaptor");
			adaptorThread.start();
		}
		if (mode == SerenityCycle.StaticAnalysis || metric == ExperimentMetric.OFF) {
			hqui.doTransaction();
			systemManager.terminateSystem();
		} else {
			for (int threadNumber = 0; threadNumber < threadLimit; threadNumber++) {
				Thread thread = new Thread(hqui);
				thread.setName(Adaptor.COMPONENT_THREAD_PREFIX + threadNumber);
				thread.start();
			}
		}
		if (mode == SerenityCycle.RunTimeContol && approach == AdaptationApproach.SERENITY
				&& metric != ExperimentMetric.OFF) {
			Thread.sleep(WARM_UP_TIME);
			adaptor.requestForAdaptation();
			Thread.sleep(500);
			adaptor.adaptComponent();
		}
	}

	private static HQUI init(ComponentBasedSystemManager systemManager, SerenityCycle mode,
			AdaptationApproach approach, double lambda, int threadLimit) throws Exception {
		adaptor = Adaptor.getInstance(systemManager, approach);
		interArrivalTime = getPoissonDistArrivalTime(lambda, threadLimit);
		runningThreads = new ArrayList<String>();
		HQUI hqui = (HQUI) context.getBean("hqui");
		startTime = System.currentTimeMillis();
		systemManager.initSystem(mode, approach);
		hqui.init();
		lastRequestArrivalTime = startTime;
		return hqui;
	}

	private static long[] getPoissonDistArrivalTime(double lambda, int threadLimit) {
		long[] arrivalTime = new long[threadLimit];
		arrivalTime[0] = 1000;
		for (int i = 1; i < threadLimit; i++) {
			long interArrival = (long) ((-Math.log(1.0 - Math.random()) / lambda) * 1000);
			arrivalTime[i] = arrivalTime[i - 1] + interArrival;
		}
		return arrivalTime;
	}

	public enum SerenityCycle {
		StaticAnalysis("static"), RunTimeContol("runtime");
		SerenityCycle(String mode) {
			this.mode = mode;
		}

		String mode;
	}

	public enum AdaptationApproach {
		SERENITY("serenity"), TRANQUILITY("tranquility");
		AdaptationApproach(String approach) {
			this.approach = approach;
		}

		String approach;
	}

	public enum TestScenario {
		ANALYZER("analyzer"), ADVISOR("advisor"), SIMULATOR("simulator");
		TestScenario(String name) {
			this.name = name;
		}

		String name;
	}

	public enum ExperimentMetric {
		REACHABILITY("reachability"), DISRUPTION("disruption"), OFF("off");
		ExperimentMetric(String name) {
			this.name = name;
		}

		String name;
	}

	private static class JVMShutdownHook extends Thread {
		public void run() {
			Adaptor.log.fatal(" D:" + (System.currentTimeMillis() - startTime) + "\n");
		}
	}
}