package edu.uci.ics.seal.savasana.runtime_old;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.After;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Before;
import org.aspectj.lang.annotation.Pointcut;

import edu.uci.ics.seal.savasana.staticAnalyisis.ComponentDependencyModelExtractor;
import edu.uci.ics.seal.savasana.staticAnalyisis.model.IndependentTransaction;
import edu.uci.ics.seal.savasana.staticAnalyisis.model.Registry;
import edu.uci.ics.seal.savasana.benchmark.systemUnderTestInterfaces.ComponentBasedSystemManager;
import edu.uci.ics.seal.savasana.runtime_old.Main_Old.AdaptationApproach;
import edu.uci.ics.seal.savasana.runtime_old.Main_Old.ExperimentMetric;
import edu.uci.ics.seal.savasana.runtime_old.Main_Old.SerenityCycle;

@Aspect
public class Monitor {

	private IndependentTransaction firstUnit;
	private Map<Long, IndependentTransaction> currentUnitMap;
	private String targetClass;
	private Adaptor adaptor;
	private boolean monitorSwitch;
	private SerenityCycle mode;
	private Registry registry;

	@Pointcut("@within(edu.uci.ics.seal.savasana.systemUnderTestInterfaces.AdaptableComponent)")
	public void beanAnnotatedWithAdaptableComponent() {
	}

	@Pointcut("execution(public * *(..))")
	public void publicMethod() {
	}

	@Pointcut("publicMethod() && beanAnnotatedWithAdaptableComponent()")
	public void publicMethodInsideAClassMarkedWithAdaptableComponent() {
	}

	@After("publicMethodInsideAClassMarkedWithAdaptableComponent()")
	public void doAfterTargetComponentExec(JoinPoint joinPoint) {
		if (!monitorSwitch)
			return;
		if (mode == SerenityCycle.StaticAnalysis)
			checkForAdaptationByStaticAnalysis(joinPoint);
		else
			checkForAdaptationByRegistry(joinPoint);
	}

	@Before("publicMethodInsideAClassMarkedWithAdaptableComponent()")
	public void doBeforeTargetComponentExec(JoinPoint joinPoint) {
		if (!monitorSwitch)
			return;
		if (mode != SerenityCycle.StaticAnalysis)
			return;
		Long threadId = Thread.currentThread().getId();
		if (!currentUnitMap.containsKey(threadId))
			currentUnitMap.put(threadId, firstUnit);
		IndependentTransaction currentTransaction = currentUnitMap.get(threadId);
		if (!currentTransaction.getEntryCall().equalsIgnoreCase(joinPoint.getSignature().getName()))
			return;
	}

	@After("execution(* edu.uci.ics.seal.savasana.systemUnderTestInterfaces.ComponentBasedSystemManager.initSystem(..))")
	public void init(JoinPoint joinPoint) {
		try {
			ComponentBasedSystemManager cbsm = (ComponentBasedSystemManager) joinPoint.getThis();
			Properties prop = cbsm.getSystemConfig();
			monitorSwitch = Boolean.valueOf(prop.getProperty("monitoring"));
			if (cbsm.getApproach() == AdaptationApproach.TRANQUILITY || Main_Old.metric == ExperimentMetric.OFF)
				monitorSwitch = false;
			if (!monitorSwitch)
				return;
			mode = cbsm.getMode();
			String targetClass = prop.getProperty("targetClass");
			registry = Registry.getInstance(prop.getProperty("registryAddr"));
			this.targetClass = targetClass;
			if (mode == SerenityCycle.StaticAnalysis) {
				String mainClass = prop.getProperty("mainClass");
				String targetPackage = prop.getProperty("targetPackage");
				String mainMethod = prop.getProperty("mainMethod");
				String springAppCtxFile = prop.getProperty("springAppCtxFile");
				IndependentTransaction root = ComponentDependencyModelExtractor.doIntermediateStaticAnalysis(
						targetPackage, springAppCtxFile, mainClass, mainMethod, prop);
				IndependentTransaction firstUnit = IndependentTransaction.getFirstState(root, targetClass);
				this.currentUnitMap = new HashMap<Long, IndependentTransaction>();
				this.firstUnit = firstUnit;
			} else if (mode == SerenityCycle.RunTimeContol) {
				this.adaptor = Main_Old.adaptor;
				registry.loadFromFile();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@After("execution(* edu.uci.ics.seal.savasana.systemUnderTestInterfaces.ComponentBasedSystemManager.terminateSystem(..))")
	public void terminate() {
		if (Main_Old.metric != ExperimentMetric.OFF)
			registry.saveToFile();
	}

	private void checkForAdaptationByStaticAnalysis(JoinPoint joinPoint) {
		Long threadId = Thread.currentThread().getId();
		IndependentTransaction currentTransaction = currentUnitMap.get(threadId);
		if (!currentTransaction.getEntryCall().equalsIgnoreCase(joinPoint.getSignature().getName()))
			return;
		IndependentTransaction nextTransaction = currentTransaction.getNextTransactionInTargetClass(targetClass);
		if (nextTransaction != null && nextTransaction.isSafeToAdapt()) {
			System.out.print("Monitor: *******************");
			System.out.println("It is safe to adapt the component " + joinPoint.getTarget().getClass().getSimpleName()
					+ " of " + Thread.currentThread().getName() + " after " + joinPoint.toShortString());
			registry.addToRegistry(joinPoint.toShortString());
		}
		currentUnitMap.put(threadId, nextTransaction);
	}

	private void checkForAdaptationByRegistry(JoinPoint joinPoint) {
		String point = joinPoint.toShortString();
		if (registry.isPointInRegistry(point)) {
			System.out.print("Monitor: *******************");
			System.out.println("It is safe to adapt the component " + joinPoint.getTarget().getClass().getSimpleName()
					+ " of " + Thread.currentThread().getName() + " after " + point);
			adaptor.safeToAdapt();
		}
	}
}
