package edu.uci.ics.seal.savasana.staticAnalyisis;

import java.io.File;
import java.util.Stack;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class Code2Calltrace {

	private Document callTraceDoc;
	private Stack<Element> preRootElementsStack;
	private Element rootElement;

	public Code2Calltrace() throws ParserConfigurationException {
		this.callTraceDoc = createTraceDoc();
		this.preRootElementsStack = new Stack<Element>();
		this.rootElement = createRootElement();
	}

	private Document createTraceDoc() throws ParserConfigurationException {
		DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
		Document doc = docBuilder.newDocument();
		return doc;
	}

	private Element createRootElement() {
		Element treads = callTraceDoc.createElement("Threads");
		Attr docId = callTraceDoc.createAttribute("id");
		docId.setValue("Test");
		treads.setAttributeNode(docId);
		callTraceDoc.appendChild(treads);
		Element tread = callTraceDoc.createElement("Tread");
		Attr treadId = callTraceDoc.createAttribute("id");
		treadId.setValue("1");
		tread.setAttributeNode(treadId);
		treads.appendChild(tread);
		return tread;
	}

	private void addMethod(String clazzName, String methodName, boolean isStatic) {
		Element clazz = callTraceDoc.createElement("class");
		rootElement.appendChild(clazz);
		clazz.appendChild(callTraceDoc.createCDATASection(clazzName));
		Element method = callTraceDoc.createElement("method");
		rootElement.appendChild(method);
		if (isStatic)
			method.setAttribute("static", "true");
		method.appendChild(callTraceDoc.createCDATASection(methodName));
	}

	public void addDefUse(String clazzName, String fieldName, boolean isDef) {
		Element defUse = callTraceDoc.createElement("defUse");
		defUse.setAttribute("type", isDef ? "def" : "use");
		rootElement.appendChild(defUse);
		defUse.appendChild(callTraceDoc.createCDATASection(clazzName + ":" + fieldName));
	}

	public void enterMethod(String clazz, String method, boolean isStatic, String callerName) {
		/*
		 * if(isInTheSameClass(clazz, this.rootElement)) return;
		 */
		Element callMethod = callTraceDoc.createElement("call");
		callMethod.setAttribute("callerInstance", callerName);
		rootElement.appendChild(callMethod);
		preRootElementsStack.push(rootElement);
		rootElement = callMethod;
		addMethod(clazz, method, isStatic);
	}

	public void exitMethod(String clazz) {
		/*
		 * if(isInTheSameClass(clazz, preRootElementsStack.peek())) return;
		 */
		rootElement = preRootElementsStack.pop();
	}

	public void buildResultFile(String fileName) throws TransformerException {
		TransformerFactory transformerFactory = TransformerFactory.newInstance();
		Transformer transformer = transformerFactory.newTransformer();
		transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		DOMSource source = new DOMSource(callTraceDoc);
		StreamResult result = new StreamResult((new File(fileName)).getPath());
		transformer.transform(source, result);
	}

	private boolean isInTheSameClass(String clazz, Element rootElement) {
		Element classChild = (Element) rootElement.getFirstChild();
		if (classChild != null && classChild.getNodeName().equals("class")) {
			Node classValue = classChild.getFirstChild();
			if (classValue.getNodeValue().equals(clazz))
				return true;
		}
		return false;
	}
}
