package edu.uci.ics.seal.savasana.staticAnalyisis;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.springframework.context.ApplicationContext;

import edu.uci.ics.seal.savasana.staticAnalyisis.model.BeanDefinition;
import edu.uci.ics.seal.savasana.staticAnalyisis.model.IndependentTransaction;
import soot.Body;
import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.SootMethodRef;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.jimple.InterfaceInvokeExpr;
import soot.jimple.InvokeExpr;
import soot.jimple.internal.JimpleLocalBox;
import soot.toolkits.graph.BriefUnitGraph;
import soot.toolkits.graph.UnitGraph;

public class ComponentDependencyModelExtractor {

	private String classFilter;
	private List<BeanDefinition> beanDefList;
	private IndependentTransaction currentTransaction;
	private String springAppCtxFile;
	private Properties sysConfig;

	public ComponentDependencyModelExtractor(String classFilter, String springAppCtxFile, Properties sysConfig) {
		this.classFilter = classFilter;
		this.currentTransaction = null;
		this.beanDefList = new ArrayList<BeanDefinition>();
		this.springAppCtxFile = springAppCtxFile;
		this.sysConfig = sysConfig;
	}

	public void processMethod(String clazzName, SootMethod method) {
		SootClass sClass = Scene.v().loadClassAndSupport(clazzName);
		sClass.setApplicationClass();
		Body b = method.retrieveActiveBody();
		UnitGraph graph = new BriefUnitGraph(b);
		List<Unit> heads = new ArrayList<Unit>(graph.getHeads());
		currentTransaction = new IndependentTransaction(currentTransaction, clazzName, method.getName(), graph);
		processUnit(graph, heads);
	}

	private void processUnit(UnitGraph graph, List<Unit> unvisited) {
		if (unvisited.isEmpty())
			return;
		Unit unit = unvisited.remove(0);
		checkForSpringBeanDef(unit);
		List<ValueBox> useBoxes = unit.getUseBoxes();
		for (ValueBox valueBox : useBoxes) {
			Value value = valueBox.getValue();
			if (value instanceof InvokeExpr) {
				processInvokeMethod(value);
			}
		}
		List<Unit> units = new ArrayList<Unit>(graph.getSuccsOf(unit));
		if (!units.isEmpty())
			unvisited.addAll(0, units);
		processUnit(graph, unvisited);
	}

	private void checkForSpringBeanDef(Unit unit) {
		List<ValueBox> useBoxes = unit.getUseBoxes();
		List<ValueBox> defBoxes = unit.getDefBoxes();
		for (ValueBox useBox : useBoxes) {
			if (useBox.getValue() instanceof InterfaceInvokeExpr) {
				SootMethodRef methodRef = ((InterfaceInvokeExpr) useBox.getValue()).getMethodRef();
				SootClass declaringClass = methodRef.declaringClass();
				if (declaringClass.getName().equals(ApplicationContext.class.getName())
						&& methodRef.name().equals("getBean")) {
					Value primaryDef = defBoxes.get(0).getValue();
					//TODO : map_id
					String beanId = sysConfig.getProperty(sysConfig.getProperty("targetClassId"));
					/*
					 * for (ValueBox box : useBoxes) { if (box instanceof
					 * ImmediateBox) { beanId =
					 * box.getValue().toString().replaceAll("\"", ""); break; }
					 * }
					 */
					beanDefList.add(new BeanDefinition(beanId, primaryDef, springAppCtxFile));
				}
				break;
			}
		}
		if (defBoxes != null && !defBoxes.isEmpty()) {
			for (ValueBox valueBox : useBoxes) {
				BeanDefinition.updateDefBoxes(beanDefList, valueBox.getValue(), defBoxes.get(0).getValue());
			}
		}
	}

	private void processInvokeMethod(Value value) {
		InvokeExpr r = (InvokeExpr) value;
		SootMethodRef methodRef = r.getMethodRef();
		SootClass declaringClass = methodRef.declaringClass();
		String clazz = declaringClass.getName();
		SootMethod sootMethod = methodRef.resolve();
		// boolean isStaic = methodRef.isStatic();
		if (declaringClass.getName().startsWith(classFilter)) {
			// String instanceName = "---";
			JimpleLocalBox instanceBox = null;
			List<?> methodCallUseBoxes = r.getUseBoxes();
			for (Object box : methodCallUseBoxes) {
				if (box instanceof JimpleLocalBox) {
					instanceBox = (JimpleLocalBox) box;
					// instanceName = instanceBox.getValue().toString();
				}
			}
			if (r instanceof InterfaceInvokeExpr && instanceBox != null) {
				clazz = BeanDefinition.getConcreteClass(beanDefList, instanceBox.getValue());
				sootMethod = Scene.v().loadClassAndSupport(clazz)
						.getMethod(sootMethod.getName(), sootMethod.getParameterTypes(), sootMethod.getReturnType());
			}
			processMethod(clazz, sootMethod);
		}
	}

	public IndependentTransaction getPs() {
		return currentTransaction;
	}

	public static IndependentTransaction doIntermediateStaticAnalysis(String filter, String springAppCtxFile,
			String mainClass, String mainMethod, Properties sysConfig) throws Exception {
		ComponentDependencyModelExtractor pseExtractor = new ComponentDependencyModelExtractor(filter,
				springAppCtxFile, sysConfig);
		SootClass sClass = Scene.v().loadClassAndSupport(mainClass);
		sClass.setApplicationClass();
		pseExtractor.processMethod(mainClass, sClass.getMethodByName(mainMethod));
		return pseExtractor.getPs();
	}

}
