package edu.uci.ics.seal.savasana.staticAnalyisis;

import java.util.ArrayList;
import java.util.List;

import org.springframework.context.ApplicationContext;

import edu.uci.ics.seal.savasana.staticAnalyisis.model.BeanDefinition;
import soot.Body;
import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.SootMethodRef;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.jimple.InterfaceInvokeExpr;
import soot.jimple.InvokeExpr;
import soot.jimple.internal.ImmediateBox;
import soot.jimple.internal.JInstanceFieldRef;
import soot.jimple.internal.JimpleLocalBox;
import soot.toolkits.graph.BriefUnitGraph;
import soot.toolkits.graph.UnitGraph;

public class SequenceTransformer {

	public final static String SPRING_APP_CTX_FILE = "/Users/asadeghi/Documents/workspace/HelloSoot/resources/SpringBeans.xml";

	private String classFilter;
	private Code2Calltrace code2Clalltrace;
	private List<BeanDefinition> beanDefList;

	public SequenceTransformer(String classFilter, Code2Calltrace callTracer) {
		this.classFilter = classFilter;
		this.code2Clalltrace = callTracer;
		this.beanDefList = new ArrayList<BeanDefinition>();
	}

	public void processMethod(String clazzName, SootMethod method, boolean isStatic, String instanceName) {
		code2Clalltrace.enterMethod(clazzName, method.getName(), isStatic, instanceName);
		SootClass sClass = Scene.v().loadClassAndSupport(clazzName);
		sClass.setApplicationClass();
		Body b = method.retrieveActiveBody();
		UnitGraph graph = new BriefUnitGraph(b);
		List<Unit> heads = new ArrayList<Unit>(graph.getHeads());
		processUnit(graph, heads);
		code2Clalltrace.exitMethod(clazzName);
	}

	private void processUnit(UnitGraph graph, List<Unit> unvisited) {
		if (unvisited.isEmpty())
			return;
		Unit unit = unvisited.remove(0);
		checkForSpringBeanDef(unit);
		List<ValueBox> defBoxes = unit.getDefBoxes();
		List<ValueBox> useBoxes = unit.getUseBoxes();
		processDefUseBoxes(defBoxes, true);
		processDefUseBoxes(useBoxes, false);
		for (ValueBox valueBox : useBoxes) {
			Value value = valueBox.getValue();
			if (value instanceof InvokeExpr) {
				InvokeExpr r = (InvokeExpr) value;
				SootMethodRef methodRef = r.getMethodRef();
				SootClass declaringClass = methodRef.declaringClass();
				String clazz = declaringClass.getName();
				SootMethod sootMethod = methodRef.resolve();
				boolean isStaic = methodRef.isStatic();
				if (declaringClass.getName().startsWith(classFilter)) {
					String instanceName = "---";
					JimpleLocalBox instanceBox = null;
					List methodCallUseBoxes = r.getUseBoxes();
					for (Object box : methodCallUseBoxes) {
						if (box instanceof JimpleLocalBox) {
							instanceBox = (JimpleLocalBox) box;
							instanceName = instanceBox.getValue().toString();
						}
					}
					if (r instanceof InterfaceInvokeExpr) {
						clazz = BeanDefinition.getConcreteClass(beanDefList, instanceBox.getValue());
						sootMethod = Scene.v().loadClassAndSupport(clazz).getMethod(sootMethod.getName(),
								sootMethod.getParameterTypes(), sootMethod.getReturnType());
					}
					processMethod(clazz, sootMethod, isStaic, instanceName);
				}
			}
		}
		List<Unit> units = new ArrayList<Unit>(graph.getSuccsOf(unit));
		if (!units.isEmpty())
			unvisited.addAll(0, units);
		processUnit(graph, unvisited);

	}

	private void processDefUseBoxes(List<ValueBox> valueBoxed, boolean isDef) {
		for (ValueBox valueBox : valueBoxed) {
			Value value = valueBox.getValue();
			if (value instanceof JInstanceFieldRef) {
				JInstanceFieldRef fieldRef = (JInstanceFieldRef) value;
				code2Clalltrace.addDefUse(fieldRef.getField().getDeclaringClass().getName(),
						fieldRef.getField().getName(), isDef);
			}
		}
	}

	private void checkForSpringBeanDef(Unit unit) {
		List<ValueBox> useBoxes = unit.getUseBoxes();
		List<ValueBox> defBoxes = unit.getDefBoxes();
		for (ValueBox useBox : useBoxes) {
			if (useBox.getValue() instanceof InterfaceInvokeExpr) {
				SootMethodRef methodRef = ((InterfaceInvokeExpr) useBox.getValue()).getMethodRef();
				SootClass declaringClass = methodRef.declaringClass();
				if (declaringClass.getName().equals(ApplicationContext.class.getName())
						&& methodRef.name().equals("getBean")) {
					Value primaryDef = defBoxes.get(0).getValue();
					String beanId = null;
					for (ValueBox box : useBoxes) {
						if (box instanceof ImmediateBox) {
							beanId = box.getValue().toString().replaceAll("\"", "");
							break;
						}
					}
					beanDefList.add(new BeanDefinition(beanId, primaryDef, SPRING_APP_CTX_FILE));
				}
				break;
			}
		}
		if (defBoxes != null && !defBoxes.isEmpty()) {
			for (ValueBox valueBox : useBoxes) {
				BeanDefinition.updateDefBoxes(beanDefList, valueBox.getValue(), defBoxes.get(0).getValue());
			}
		}
	}
}
