package edu.uci.ics.seal.savasana.staticAnalyisis.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import soot.Unit;
import soot.toolkits.graph.UnitGraph;

public class IndependentTransaction {

	private IndependentTransaction preTransaction;
	private String entryCall;
	private String clazz;
	private IndependentTransaction nextTransaction;
	private UnitGraph cfg;

	public IndependentTransaction(IndependentTransaction pu, String clazz, String method, UnitGraph cfg) {
		this.preTransaction = pu;
		if (pu != null)
			pu.nextTransaction = this;
		this.clazz = clazz;
		this.entryCall = method;
		this.cfg = cfg;
		this.nextTransaction = null;
	}

	public String getEntryCall() {
		return entryCall;
	}

	public String getClazz() {
		return clazz;
	}

	public IndependentTransaction getPreState() {
		return preTransaction;
	}

	public IndependentTransaction getPreTransactionInTargetClass(String clazz) {
		if (preTransaction == null)
			return null;
		if (clazz.equalsIgnoreCase(preTransaction.clazz))
			return preTransaction;
		return preTransaction.getPreTransactionInTargetClass(clazz);
	}

	public IndependentTransaction getNextTransactionInTargetClass(String clazz) {
		if (nextTransaction == null)
			return null;
		if (clazz.equalsIgnoreCase(nextTransaction.clazz))
			return nextTransaction;
		return nextTransaction.getNextTransactionInTargetClass(clazz);
	}

	private boolean matchFoundInternally(Unit targetUnit, List<Unit> startUnits) {
		DefUseItem defUse = DefUseItem.unit2DefUseItem(targetUnit);
		if (defUse == null)
			return true;
		if (defUse.isDef())
			return true;
		List<Unit> predUnits = new ArrayList<Unit>();
		predUnits.addAll(startUnits);
		Unit currentUnit = null;
		while (!predUnits.isEmpty()) {
			currentUnit = predUnits.remove(0);
			DefUseItem currentDU = DefUseItem.unit2DefUseItem(currentUnit);
			if (currentDU == null || !defUse.isCorrespondingTo(currentDU))
				predUnits.addAll(cfg.getPredsOf(currentUnit));
		}
		if (!cfg.getPredsOf(currentUnit).isEmpty())
			return true;
		return false;
	}

	private boolean matchFound(Unit targetUnit, List<Unit> startUnits, IndependentTransaction boundaryPU) {
		if (matchFoundInternally(targetUnit, startUnits))
			return true;
		IndependentTransaction pu = getPreTransactionInTargetClass(clazz);
		if (pu == null)
			return false;
		if (pu == boundaryPU)
			return false;
		return pu.matchFound(targetUnit, pu.cfg.getTails(), boundaryPU);
	}

	private boolean matchFoundForAllUnits(IndependentTransaction boundaryPS) {
		Iterator<Unit> units = cfg.iterator();
		while (units.hasNext()) {
			Unit unit = units.next();
			ArrayList<Unit> startUnits = new ArrayList<Unit>();
			startUnits.add(unit);
			if (!matchFound(unit, startUnits, boundaryPS))
				return false;
		}
		return true;
	}

	public boolean isSafeToAdapt() {
		IndependentTransaction pu = this;
		while (pu != null) {
			if (!pu.matchFoundForAllUnits(this.getPreTransactionInTargetClass(clazz)))
				return false;
			pu = pu.getNextTransactionInTargetClass(clazz);
		}
		return true;
	}

	public static boolean isSafeToAdapt(Collection<IndependentTransaction> units) {
		for (IndependentTransaction independentTransaction : units) {
			if (independentTransaction != null && !independentTransaction.isSafeToAdapt())
				return false;
		}
		return true;
	}

	public static IndependentTransaction getFirstState(IndependentTransaction lastState, String targetClass) {
		IndependentTransaction result = null;
		IndependentTransaction pu = lastState;
		if (targetClass.equalsIgnoreCase(pu.getClazz()))
			result = pu;
		while (pu.getPreState() != null) {
			pu = pu.getPreState();
			if (targetClass.equalsIgnoreCase(pu.getClazz()))
				result = pu;
		}
		return result;
	}

	@Override
	public String toString() {
		return "Transaction [entryCall=" + entryCall + ", clazz=" + clazz + "]";
	}

}